"""Module for Entity and entity Group data models"""

from os.path import join as path
from .states import State
from .base import JsonModel


class Group:
    """Represents the groups that entities belong to"""

    def __init__(self, group_id: str, client) -> None:
        """Initializes object with needed attributes"""
        self.client = client
        self.group_id = group_id
        self.entities = JsonModel()

    def __repr__(self):
        """Returns a readable string identifying each entity group."""
        return f'<Group {self.group_id}>'

    def add_entity(self, entity_slug: str, state: State) -> None:
        """Registers entities to this Group object"""
        self.entities.update({
            entity_slug: Entity(entity_slug, state, self)
        })

    def get_entity(self, entity_slug: str):
        """Returns Entity with the given name if it exists. Otherwise returns None"""
        return self.entities.get(entity_slug, None)


class Entity:
    """Represents entities inside of homeassistant"""

    def __init__(self, slug: str, state: State, group: Group) -> None:
        """Initializes object with needed attributes"""
        self.id = slug
        self.state = state
        self.group = group

    def __repr__(self) -> str:
        """Returns a readable string indentifying each Entity"""
        return f'<Entity entity_id="{self.entity_id}" state="{self.state}">'

    def get_state(self) -> State:
        """Returns the state last fetched from the api."""
        # TODO: add caching
        return self.state

    def fetch_state(self) -> State:
        """Asks homeassistant for the state of the entity and sets it locally"""
        state_data = self.group.client.request(path(
            'states',
            self.entity_id
        ))
        self.state = self.group.client.process_state_json(state_data)
        return self.state

    def set_state(self, state: State) -> State:
        """Tells homeassistant to set the given State object (you can construct the state object yourself)"""
        state_data = self.group.client.request(
            path(
                'states',
                self.group.group_id + '.' + self.id
            ),
            method='POST',
            json=state
        )
        self.state = self.group.client.process_state_json(state_data)
        return self.state

    @property
    def entity_id(self):
        """Constructs the entity_id string from its group and slug"""
        return self.group.group_id + '.' + self.id



from .base import JsonModel
from ..errors import ParameterMissingError


class State(JsonModel):
    """A class representing a state of an entity."""

    def __init__(self, json: dict = None, **data):
        super().__init__(json, **data)
        if 'entity_id' not in self:
            raise ParameterMissingError('"entity_id" attribute is a required parameter')
        if 'state' not in self:
            raise ParameterMissingError('"state" attribute is a required parameter')

        self.update(
            attributes=self.get('attributes', {}),
            last_changed=self.get('last_changed', None),
            last_updated=self.get('last_updated', None),
            context=self.get('context', {})
        )

    def __repr__(self):
        return f'<EntityState "{self.state}">'



"""File for base data model types"""

from typing import Any


class JsonModel(dict):
    """Makes dict values available as object attributes"""
    def __init__(self, json: dict = None, **kwargs) -> None:
        """Updates self with passed json data and kwargs"""
        if json:
            self.update(json)
        if kwargs:
            self.update(kwargs)

    def __getattr__(self, name: str) -> Any:
        """Searches for a key name in self if not found returns from parent attributes"""
        if '__' not in str(name):
            if name in self:
                if isinstance(self[name], dict):
                    return JsonModel(self[name])
                return self[name]
        return super().__getattribute__(name)

    def __setattr__(self, name: str, value: Any) -> None:
        """Sets values as dictionary entries if not dunder methods or attributes"""
        if '__' not in str(name):
            if isinstance(value, dict):
                self[name] = JsonModel(value)
            self[name] = value
        super().__setattr__(name, value)



"""Module for parent RawWrapper class"""

import os
import json
import simplejson
import requests
from typing import Union

from .errors import MalformedDataError


class RawWrapper:
    """Builds, and makes requests to the API"""

    def __init__(self, api_url: str, token: str) -> None:
        """Prepares and stores API URL and Love Lived Access Token token"""
        self.api_url = api_url
        if not self.api_url.endswith('/'):
            self.api_url += '/'
        self._token = token

    def endpoint(self, path: str) -> str:
        """Joins the api base url with a local path to an absolute url"""
        url = os.path.join(self.api_url, path)
        return url

    @property
    def _headers(self) -> dict:
        """Constructs the headers to send to the api for every request"""
        return {
            "Authorization": f"Bearer {self._token}",
            "content-type": "application/json",
        }

    def request(
        self,
        path,
        method='GET',
        headers: dict = None,
        return_text_if_fail=False,
        **kwargs
    ) -> Union[dict, list, str]:
        """Base method for making requests to the api"""
        if headers is None:
            headers = {}
        if isinstance(headers, dict):
            headers.update(self._headers)
        else:
            raise ValueError(f'headers must be dict or dict subclass, not type "{type(headers).__name__}"')

        resp = requests.request(
            method,
            self.endpoint(path),
            headers=headers,
            **kwargs
        )
        return self.response_logic(resp, return_text_if_fail)

    def response_logic(self, response: requests.Response, return_text_if_fail=False) -> Union[dict, list, str]:
        """Processes reponses from the api and formats them"""
        try:
            res = response.json()
        except (json.decoder.JSONDecodeError, simplejson.decoder.JSONDecodeError):
            if return_text_if_fail:
                return response.text
            else:
                raise MalformedDataError(f'Homeassistant responded with non-json response: {repr(response.text)}')
        else:
            return res

    def construct_params(self, params: dict) -> str:
        """Custom method for constructing non-standard query strings"""
        return '&'.join([
            k if v is None
            else f"{k}={v}"
            for k, v in params.items()
        ])



"""Module for custom error classes"""


class Error(BaseException):
    """Base class for custom errors"""


class MalformedDataError(Error):
    """Error raised when data from api is not formatted as JSON"""


class MalformedInputError(Error):
    """Error raised when user passes malformed data in parameters"""


class ResponseError(Error):
    """Error raised when an api reponse is formatted unexpectedly."""


class APIConfigurationError(Error):
    """Error raised when api says it has an invalid configuration file"""


class ParameterMissingError(Error):
    """Error raised when an expected attribute is missing from api response data"""



"""Module for Client class"""

from datetime import datetime
from os.path import join as path
from typing import List

from .models import Group, Entity, State, Domain, JsonModel, Event
from .errors import APIConfigurationError, ResponseError
from .rawapi import RawWrapper


class RawClient(RawWrapper):
    """The class used to interact with all of the api"""

    def __init__(self, *args, **kwargs) -> None:
        """Passes API authentication data to RawWrapper and validates API status"""
        super().__init__(*args, **kwargs)
        self.check_api_running()
        self.check_api_config()

    def __repr__(self) -> str:
        """Returns string to intuitively indentify Client classes from `repr()`"""
        return f'<Client of "{self.api_url[:20]}">'

    # Response processing methods
    def process_services_json(self, json: dict) -> Domain:
        """Constructs Domain and Service models from json data"""

        domain = Domain(json.get('domain'), self)
        for service_id, data in json.get('services').items():
            domain.add_service(service_id, **data)
        return domain

    def process_state_json(self, json: dict) -> State:
        """Constructs State model from json data"""
        return State(**json)

    def process_event_json(self, json: dict) -> Event:
        """Constructs Event model from json data"""
        return Event(**json)

    # API information methods
    def api_error_log(self) -> str:
        """Returns the server error log as a string"""
        return self.request('error_log', return_text_if_fail=True)

    def api_config(self) -> dict:
        """Returns the yaml configuration of homeassistant"""
        res = self.request('config')
        return res

    def logbook_entries(
        self,
        entity: Entity = None,
        start_date: datetime = None,  # Defaults to 1 day before
        end_date: datetime = None
    ):
        raise NotImplementedError('Not implemented in this pre-release as of 2.0a1')

    def get_history(
        self,
        entities: tuple = None,
        entity: Entity = None,
        start_date: datetime = None,  # Defaults to 1 day before
        end_date: datetime = None,
        minimal_state_data=False,
        big_changes_only=False
    ):
        raise NotImplementedError('Not implemented in this pre-release as of 2.0a1')

    def get_rendered_template(self, template: str):
        raise NotImplementedError('Not implemented in this pre-release as of 2.0a1')

    def get_discovery_info(self) -> dict:
        """Returns a dictionary of discovery info such as internal_url and version"""
        res = self.request('discovery_info')
        return res

    # API check methods
    def check_api_config(self) -> True:
        """Asks homeassistant to validate its configuration file"""
        res = self.request('config/core/check_config', method='POST')
        valid = {'valid': True, 'invalid': False}.get(res['result'], None)
        if not valid:
            raise APIConfigurationError(res['errors'])
        return valid

    def check_api_running(self) -> True:
        """Asks homeassistant if its running"""
        res = self.request('')
        if res.get('message', None) == 'API running.':
            return True
        else:
            raise ResponseError('Server response did not return message attribute')

    def malformed_id(self, entity_id: str) -> bool:
        """Checks whether or not a given entity_id is formatted correctly"""
        checks = [
            ' ' in entity_id,
            '.' not in entity_id,
            '-' in entity_id,
            entity_id.lower() == entity_id
        ]
        return True in checks

    # Entity methods
    def get_entities(self) -> JsonModel:
        """Fetches all entities from the api"""
        class GroupDict(dict):
            """dict subclass for constructing dynamic default Group models"""
            def __missing__(cls, group_id: str):
                """Allows for dynamic default values in a dictionary"""
                cls[group_id] = Group(group_id, self)
                return cls[group_id]
        entities = GroupDict()
        for state in self.get_states():
            group_id, entity_slug = state.entity_id.split('.')
            entities[group_id].add_entity(entity_slug, state)
        return JsonModel(entities)

    def get_entity(
        self,
        group_id: str = None,
        entity_slug: str = None,
        entity_id: str = None
    ) -> Entity:
        """Returns a Entity model for an entity_id"""
        if group_id is not None and entity_slug is not None:
            state = self.get_state(group=group_id, slug=entity_slug)
        elif entity_id is not None:
            state = self.get_state(entity_id=entity_id)
        else:
            raise ValueError('Neither group and slug or entity_id provided. {help_msg}'.format(
                help_msg='Use keyword arguments to pass entity_id. Or you can pass the entity_group and entity_slug instead'
            ))
        group_id, entity_slug = state.entity_id.split('.')
        group = Group(group_id, self)
        group.add_entity(entity_slug, state)
        return group.get_entity(entity_slug)

    # Services and domain methods
    def get_domains(self) -> JsonModel:
        """Fetches all Services from the api"""
        services = self.request('services')
        services = [self.process_services_json(data) for data in services]
        services = {service.domain_id: service for service in services}
        return JsonModel(services)

    def trigger_service(
        self,
        domain: str,
        service: str,
        **service_data
    ) -> List[State]:
        """Tells homeassistant to trigger a service, returns stats changed while being called"""
        data = self.request(
            path(
                'services',
                domain,
                service
            ),
            method='POST',
            json=service_data
        )
        return [
            self.process_state_json(state_data)
            for state_data in data
        ]

    # EntityState methods
    def get_state(self, entity_id: str = None, group: str = None, slug: str = None) -> State:
        """Fetches the state of the entity specified"""
        if group is not None and slug is not None:
            entity_id = group + '.' + slug
        elif entity_id is None:
            raise ValueError('Neither group and slug or entity_id provided.')
        data = self.request(path('states', entity_id))
        return self.process_state_json(data)

    def set_state(self, entity_id: str = None, state: str = None, group: str = None, slug: str = None, **payload) -> State:
        """Sets the state of the entity given (does not have to be a real entity) and returns the updated state"""
        if group is None or slug is None:
            raise ValueError('To use group or slug you need to pass both not just one.'
                             'Make sure you are using keyword arguments.')
        if group is not None and slug is not None:
            entity_id = group + '.' + slug
        elif entity_id is None:
            raise ValueError('Neither group and slug or entity_id provided.')
        if state is None:
            raise ValueError('required parameter "state" is missing')
        payload.update(state=state)
        data = self.request(
            path('states', entity_id),
            method='POST',
            json=payload
        )
        return self.process_state_json(data)

    def get_states(self) -> List[State]:
        """Gets the states of all entitites within homeassistant"""
        data = self.request('states')
        return [self.process_state_json(state_data) for state_data in data]

    # Event methods
    def get_events(self) -> JsonModel:
        """Gets the Events that happen within homeassistant"""
        data = self.request('events')
        events = [self.process_event_json(event_info) for event_info in data]
        events = {evt.event_type: evt for evt in events}
        return JsonModel(events)

    def fire_event(self, event_type: str, **event_data) -> str:
        """Fires a given event_type within homeassistant. Must be an existing event_type."""
        data = self.request(
            path(
                'events',
                event_type
            ),
            method='POST',
            json=event_data
        )
        return data.get('message', 'No message provided')


class Client(RawClient):
    """Uses RawClient to integrate data models as parameters"""



"""File for Service and Domain data models"""

from os.path import join as path
from typing import List

from .base import JsonModel
from .states import State


class Domain:
    """A class representing the domain that services belong to."""

    def __init__(self, domain: str, client) -> None:
        """Initializes needed attributes"""
        self.domain_id = domain
        self.client = client
        self.services = JsonModel()

    def __repr__(self) -> str:
        """Returns readable string indentifying each Domain class"""
        return f'<Domain {self.domain_id}>'

    def add_service(self, service_id: str, **data) -> None:
        """Registers services into a domain to be used or accessed"""
        self.services.update({
            service_id: Service(service_id, self, **data)
        })

    def get_service(self, service_id: str):
        """Return a Service with the given service_id, returns None if no such service exists"""
        return self.services.get(service_id, None)

    def __getattr__(self, attr: str):
        """Allows services accessible as attributes"""
        if hasattr(self, attr):
            return super().__getattribute__(attr)
        if attr in self.services:
            return self.get_service(attr)
        return super().__getattribute__(attr)


class Service:
    """Class representing services from homeassistant"""

    def __init__(
        self,
        service_id: str,
        domain: Domain,
        name: str = None,
        description: str = None,
        fields: dict = None,
        target: dict = None
    ) -> None:
        self.id = service_id
        self.domain = domain
        self.name = name
        self.description = description
        self.fields = fields
        self.target = target

    def __repr__(self):
        """Returns a readable string indentifying each Service"""
        return f'<Service {self.id} domain="{self.domain.domain_id}">'

    def trigger(self, **service_data) -> List[State]:
        """Triggers the service associated with this object."""

        data = self.domain.client.request(
            path(
                'services',
                self.domain.domain_id,
                self.id
            ),
            method='POST',
            json=service_data
        )
        return [
            self.domain.client.process_state_json(state_data)
            for state_data in data
        ]



from setuptools import setup
from homeassistant_api import __version__

with open("README.md", "r") as f:
    read = f.read()

setup(
    name="HomeAssistant API",
    url="https://github.com/GrandMoff100/HomeassistantAPI",
    description="Python Wrapper for Homeassistant's REST API",
    version=__version__,
    keywords=['homeassistant', 'api', 'wrapper', 'client'],
    author="GrandMoff100",
    author_email="nlarsen23.student@gmail.com",
    packages=[
        "homeassistant_api",
        "homeassistant_api.models",
        # "homeassistant_api._async",
        # "homeassistant_api._async.models"
    ],
    long_description=read,
    long_description_content_type="text/markdown",
    install_requires=["requests", "simplejson"],
    extras_require={
        "async": ["aiohttp"]
    },
    python_requires=">=2.7",
    provides=["homeassistant_api"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Version Control :: Git"
    ]
)



"""Event Model File"""
from os.path import join as path


class Event:
    """
    Event class for Homeassistant Event Triggers

    For attribute information see the Data Science docs on Event models
    https://data.home-assistant.io/docs/events
    """

    def __init__(self, event: str, listener_count: int, client):
        self.event_type = event
        self.listener_count = listener_count
        self.client = client

    def __repr__(self):
        return f'<Event {self.event_type}>'

    def fire(self, **event_data):
        data = self.client.request(
            path(
                'events',
                self.event_type
            ),
            method='POST',
            json=event_data
        )
        return data.get('message', 'No message provided')
